"""
    is_entry_node(node::Node)

Return whether this node is an entry node in its graph, i.e., it has no children.
"""
is_entry_node(node::Node) = length(node.children) == 0

"""
    is_exit_node(node::Node)

Return whether this node is an exit node of its graph, i.e., it has no parents.
"""
is_exit_node(node::Node) = length(node.parents) == 0

"""
    data(edge::Edge)

Return the data transfered by this edge, i.e., 0 if the child is a [`ComputeTaskNode`](@ref), otherwise the child's `data()`.
"""
function data(edge::Edge)
    if typeof(edge.edge[1]) <: DataTaskNode
        return data(edge.edge[1].task)
    end
    return 0.0
end

"""
    children(node::Node)

Return a copy of the node's children so it can safely be muted without changing the node in the graph.

A node's children are its prerequisite nodes, nodes that need to execute before the task of this node.
"""
function children(node::Node)
    return copy(node.children)
end

"""
    parents(node::Node)

Return a copy of the node's parents so it can safely be muted without changing the node in the graph.

A node's parents are its subsequent nodes, nodes that need this node to execute.
"""
function parents(node::Node)
    return copy(node.parents)
end

"""
    siblings(node::Node)

Return a vector of all siblings of this node. 

A node's siblings are all children of any of its parents. The result contains no duplicates and includes the node itself.
"""
function siblings(node::Node)
    result = Set{Node}()
    push!(result, node)
    for parent in node.parents
        union!(result, parent.children)
    end

    return result
end

"""
    partners(node::Node)

Return a vector of all partners of this node. 

A node's partners are all parents of any of its children. The result contains no duplicates and includes the node itself.

Note: This is very slow when there are multiple children with many parents. 
This is less of a problem in [`siblings(node::Node)`](@ref) because (depending on the model) there are no nodes with a large number of children, or only a single one.
"""
function partners(node::Node)
    result = Set{Node}()
    push!(result, node)
    for child in node.children
        union!(result, child.parents)
    end

    return result
end

"""
    partners(node::Node, set::Set{Node})

Alternative version to [`partners(node::Node)`](@ref), avoiding allocation of a new set. Works on the given set and returns `nothing`.
"""
function partners(node::Node, set::Set{Node})
    push!(set, node)
    for child in node.children
        union!(set, child.parents)
    end
    return nothing
end

"""
    is_parent(potential_parent::Node, node::Node)

Return whether the `potential_parent` is a parent of `node`.
"""
function is_parent(potential_parent::Node, node::Node)
    return potential_parent in node.parents
end

"""
    is_child(potential_child::Node, node::Node)

Return whether the `potential_child` is a child of `node`.
"""
function is_child(potential_child::Node, node::Node)
    return potential_child in node.children
end
