import MetagraphOptimization.ABCParticle

using QEDbase

include("../examples/profiling_utilities.jl")

@testset "Unit Tests Execution" begin
    machine = get_machine_info()

    process_2_2 = ABCProcessDescription(
        Dict{Type, Int64}(ParticleA => 1, ParticleB => 1),
        Dict{Type, Int64}(ParticleA => 1, ParticleB => 1),
    )

    particles_2_2 = ABCProcessInput(
        process_2_2,
        ABCParticle[
            ParticleA(SFourMomentum(0.823648, 0.0, 0.0, 0.823648)),
            ParticleB(SFourMomentum(0.823648, 0.0, 0.0, -0.823648)),
        ],
        ABCParticle[
            ParticleA(SFourMomentum(0.823648, -0.835061, -0.474802, 0.277915)),
            ParticleB(SFourMomentum(0.823648, 0.835061, 0.474802, -0.277915)),
        ],
    )
    expected_result = 0.00013916495566048735

    @testset "AB->AB no optimization" begin
        for _ in 1:10   # test in a loop because graph layout should not change the result
            graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->AB.txt"), ABCModel())
            @test isapprox(execute(graph, process_2_2, machine, particles_2_2), expected_result; rtol = 0.001)

            # graph should be fully scheduled after being executed
            @test is_scheduled(graph)

            func = get_compute_function(graph, process_2_2, machine)
            @test isapprox(func(particles_2_2), expected_result; rtol = 0.001)
        end
    end

    @testset "AB->AB after random walk" begin
        for i in 1:1000
            graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->AB.txt"), ABCModel())
            random_walk!(graph, 50)

            @test is_valid(graph)

            @test isapprox(execute(graph, process_2_2, machine, particles_2_2), expected_result; rtol = 0.001)

            # graph should be fully scheduled after being executed
            @test is_scheduled(graph)
        end
    end

    process_2_4 = ABCProcessDescription(
        Dict{Type, Int64}(ParticleA => 1, ParticleB => 1),
        Dict{Type, Int64}(ParticleA => 1, ParticleB => 3),
    )
    particles_2_4 = gen_process_input(process_2_4)
    graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->ABBB.txt"), ABCModel())
    expected_result = execute(graph, process_2_4, machine, particles_2_4)

    @testset "AB->ABBB no optimization" begin
        for _ in 1:5   # test in a loop because graph layout should not change the result
            graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->ABBB.txt"), ABCModel())
            @test isapprox(execute(graph, process_2_4, machine, particles_2_4), expected_result; rtol = 0.001)

            func = get_compute_function(graph, process_2_4, machine)
            @test isapprox(func(particles_2_4), expected_result; rtol = 0.001)
        end
    end

    @testset "AB->ABBB after random walk" begin
        for i in 1:200
            graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->ABBB.txt"), ABCModel())
            random_walk!(graph, 100)
            @test is_valid(graph)

            @test isapprox(execute(graph, process_2_4, machine, particles_2_4), expected_result; rtol = 0.001)
        end
    end

    @testset "AB->AB large sum fusion" for _ in 1:20
        graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->AB.txt"), ABCModel())

        # push a fusion with the sum node
        ops = get_operations(graph)
        for fusion in ops.nodeFusions
            if isa(fusion.input[3].task, ComputeTaskSum)
                push_operation!(graph, fusion)
                break
            end
        end

        # push two more fusions with the fused node
        for _ in 1:15
            ops = get_operations(graph)
            for fusion in ops.nodeFusions
                if isa(fusion.input[3].task, FusedComputeTask)
                    push_operation!(graph, fusion)
                    break
                end
            end
        end

        # try execute
        @test is_valid(graph)
        expected_result = 0.00013916495566048735
        @test isapprox(execute(graph, process_2_2, machine, particles_2_2), expected_result; rtol = 0.001)
    end


    @testset "AB->AB large sum fusion" for _ in 1:20
        graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->AB.txt"), ABCModel())

        # push a fusion with the sum node
        ops = get_operations(graph)
        for fusion in ops.nodeFusions
            if isa(fusion.input[3].task, ComputeTaskSum)
                push_operation!(graph, fusion)
                break
            end
        end

        # push two more fusions with the fused node
        for _ in 1:15
            ops = get_operations(graph)
            for fusion in ops.nodeFusions
                if isa(fusion.input[3].task, FusedComputeTask)
                    push_operation!(graph, fusion)
                    break
                end
            end
        end

        # try execute
        @test is_valid(graph)
        expected_result = 0.00013916495566048735
        @test isapprox(execute(graph, process_2_2, machine, particles_2_2), expected_result; rtol = 0.001)
    end

    @testset "AB->AB fusion edge case" for _ in 1:20
        graph = parse_dag(joinpath(@__DIR__, "..", "input", "AB->AB.txt"), ABCModel())

        # push two fusions with ComputeTaskV
        for _ in 1:2
            ops = get_operations(graph)
            for fusion in ops.nodeFusions
                if isa(fusion.input[1].task, ComputeTaskV)
                    push_operation!(graph, fusion)
                    break
                end
            end
        end

        # push fusions until the end
        cont = true
        while cont
            cont = false
            ops = get_operations(graph)
            for fusion in ops.nodeFusions
                if isa(fusion.input[1].task, FusedComputeTask)
                    push_operation!(graph, fusion)
                    cont = true
                    break
                end
            end
        end

        # try execute
        @test is_valid(graph)
        expected_result = 0.00013916495566048735
        @test isapprox(execute(graph, process_2_2, machine, particles_2_2), expected_result; rtol = 0.001)
    end

end
println("Execution Unit Tests Complete!")
